from . import Enum

class SegySampleFormat(Enum):
    IBM_FLOAT_4_BYTE = 1
    SIGNED_INTEGER_4_BYTE = 2
    SIGNED_SHORT_2_BYTE = 3
    FIXED_POINT_WITH_GAIN_4_BYTE = 4
    IEEE_FLOAT_4_BYTE = 5
    IEEE_FLOAT_8_BYTE = 6
    SIGNED_CHAR_3_BYTE = 7
    SIGNED_CHAR_1_BYTE = 8
    SIGNED_INTEGER_8_BYTE = 9
    UNSIGNED_INTEGER_4_BYTE = 10
    UNSIGNED_SHORT_2_BYTE = 11
    UNSIGNED_INTEGER_8_BYTE = 12
    UNSIGNED_INTEGER_3_BYTE = 15
    UNSIGNED_CHAR_1_BYTE = 16
    NOT_IN_USE_1 = 19
    NOT_IN_USE_2 = 20
